/**
 * P4
 * http://snarfed.org/space/p4
 *
 * Copyright 2001-2004
 * Ken Ashcraft <kashcraft+p4@gmail.com>
 * Ryan Barrett <p4@ryanb.org>
 * Maulik Shah <p4@maulik.net>
 * Nathan Stoll <nathanstoll+p4@gmail.com>
 *
 * @file debug.h
 *
 * Defines the cdebug class, a general-purpose class designed to support
 * debugging output with iostreams. The cdebug class is derived from ostream,
 * so it implements the stream insertion operator (<<).
 *
 * Debugging output can be turned on and off easily. Before it prints anything,
 * cdebug checks the return value from the is_on method, and only prints the
 * debugging statement if is_on returns true. This method can be implemented
 * any one of a number of ways - right now it uses csettings to check whether
 * the debug_output setting is true or false.
 *
 * Also, cdebug supports fine-grained control of debugging statements with
 * module settings. Before inserting anything into the cdebug stream, you must
 * insert a call to the parameterized manipulator acquire. This manipulator
 * takes a string, which is the name of a setting (called a module setting).
 * This setting is retrieved from csettings. If it's true, the debugging
 * statement is printed; otherwise, the stream insertion operator does nothing.
 *
 * For example, this debugging output will print "module1: hello 1.23" if the
 * setting "module1" is true. Otherwise, nothing will be printed.
 *
 *  cdebug::instance() << acquire("module1") << "hello " << 1.23 << release();
 *
 * The cdebug class is implemented as a singleton. Use the instance method to
 * retrieve the cdebug instance.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */

#ifndef _DEBUG_H
#define _DEBUG_H

#include "settings.h"

#include <pthread.h>
#include <string>
#include <iostream>
#include <fstream>
using namespace std;


/*****************************************************************************
 * FORWARD DECLARATIONS
 *****************************************************************************/
class cdebug;
class acquire;
class release;
cdebug &operator<< (cdebug &debug, const acquire &a);
cdebug &operator<< (cdebug &debug, const release &r);


/*****************************************************************************
 * CDEBUG CLASS
 *****************************************************************************/
class cdebug
{
  public:

  /** instance
   *
   * Returns the only instance of cdebug. This implements the singleton
   * pattern.
   */
  static cdebug &instance();

  /** operator<<
   *
   * Stream insertion operator. If the debug stream is currently owned (as a
   * result of the acquire manipulator), inserts into cerr. Otherwise, does
   * nothing.
   */
  template <typename T>
  friend cdebug &operator<< (cdebug &debug, const T &elem);


  protected:
  // see below for acquire and release classes
  friend cdebug &operator<< (cdebug &debug, const acquire &a);
  friend cdebug &operator<< (cdebug &debug, const release &a);

  /** ctors, dtor
   */
  cdebug();
  cdebug(const string &filename);
  ~cdebug();

  /** is_on
   *
   * Returns true if debugging output for the given module is turned on, false
   * if it's turned off. (If the master debug setting is turned off, is_off
   * always returns false.)
   */
  bool is_on(const string &module) const;


  /**************************************************************************
   * CONSTANTS AND MEMBERS
   **************************************************************************/
  protected:
  pthread_mutex_t mtx;
  string module;  // module that currently owns cdebug; protected by mtx

  // !!! IMPORTANT !!! the filename ctor initializes out to refer to file, so
  // file *must* be declared before out here.
  fstream file;   // only opened if a log file is specified
  ostream &out;   // points to the log file if specified, cerr otherwise

  public:
  static const string kmaster_setting;
  static const string klogfile_setting;
};

/*****************************************************************************
 * ACQUIRE CLASS
 *****************************************************************************/
/** acquire
 *
 * A parameterized manipulator that takes the module name to print the
 * debugging statement under. The module name is a string that must be a
 * setting currently loaded into csettings. If that setting is true (1),
 * the debugging statement will be printed; otherwise, nothing will be
 * printed. (The default module name is the global debug setting.)
 *
 * If is_on() returns false, acquire does nothing.
 *
 * A call to the acquire manipulator is valid until the release manipulator
 * is called.
 *
 * NOTE: the acquire manipulator is *blocking* - if the module setting is
 * true, and another thread currently owns the cdebug instance, acquire
 * blocks until that thread releases the mutex.
 */
class acquire
{
  public:
  acquire(const string &module = cdebug::kmaster_setting) : module(module) {}
  ~acquire() {}
  friend cdebug &operator<< (cdebug &debug, const acquire &a);

  protected:
  string module;
};


/*****************************************************************************
 * RELEASE CLASS
 *****************************************************************************/
/** release
 *
 * A manipulator that releases the cdebug instance - should only be called
 * after a call to acquire.
 */
class release
{
  public:
  release()  {}
  ~release()  {}
  friend cdebug &operator<< (cdebug &debug, const release &r);
};


/*****************************************************************************
 * TEMPLATIZED METHODS
 *****************************************************************************/
template <typename T>
cdebug &operator<< (cdebug &debug, const T &elem)
{
  if (!debug.is_on(debug.module))
    return debug;  // this debug module is turned off

  debug.out << elem;
  return debug;
}


/*****************************************************************************
 * INLINE METHODS
 *****************************************************************************/
// Note that, yes, there are actually *two* instances of cdebug, one that
// points to cerr and one that points to the log file, if any. only one is ever
// returned and used, though. 
inline cdebug &cdebug::instance()
{
  // by default, cdebug writes to cerr....
  static cdebug debug_err;

  // ...but if a log file is specified, use that instead
  static string filename = csettings::instance().get<string>(klogfile_setting);
  static cdebug debug_file(filename);

  if (filename == "")
    return debug_err;
  else
    return debug_file;
}


inline bool cdebug::is_on(const string &module) const
{
  return (csettings::instance().get<bool>(kmaster_setting) &&
          module != "" && csettings::instance().get<bool>(module));
}

#endif  // _DEBUG_H


